@@ -26,6 +26,8 @@
 #include <sys/stat.h>
 #include <sys/wait.h>
 
+USES_APPLE_DEPRECATED_API
+
 #if defined(HAVE_MALLOPT) && defined(HAVE_MALLOC_H)
 #  include <malloc.h>
 #endif
@@ -599,7 +601,7 @@ int fr_set_dumpable(bool allow_core_dumps)
 		struct rlimit no_core;
 
 		no_core.rlim_cur = 0;
-		no_core.rlim_max = 0;
+		no_core.rlim_max = core_limits.rlim_max;
 
 		if (setrlimit(RLIMIT_CORE, &no_core) < 0) {
 			fr_strerror_printf("Failed disabling core dumps: %s", fr_syserror(errno));
@@ -1177,7 +1179,7 @@ void NEVER_RETURNS _fr_exit(char const *file, int line, int status)
 #ifndef NDEBUG
 	char const *error = fr_strerror();
 
-	if (error && (status != 0)) {
+	if (error && *error && (status != 0)) {
 		FR_FAULT_LOG("EXIT(%i) CALLED %s[%u].  Last error was: %s", status, file, line, error);
 	} else {
 		FR_FAULT_LOG("EXIT(%i) CALLED %s[%u]", status, file, line);
