@@ -15,7 +15,7 @@
  */
 
 /**
- * $Id: 6228438c47f9f5941e8a0535f44dc152d1b04414 $
+ * $Id: b0b3e24028c9d634b1a4d51b3b0e3d2ffd8798e4 $
  * @file main/client.c
  * @brief Manage clients allowed to communicate with the server.
  *
@@ -24,7 +24,7 @@
  * @copyright 2000 Alan DeKok <aland@ox.org>
  * @copyright 2000 Miquel van Smoorenburg <miquels@cistron.nl>
  */
-RCSID("$Id: 6228438c47f9f5941e8a0535f44dc152d1b04414 $")
+RCSID("$Id: b0b3e24028c9d634b1a4d51b3b0e3d2ffd8798e4 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/rad_assert.h>
@@ -41,11 +41,16 @@ RCSID("$Id: 6228438c47f9f5941e8a0535f44dc152d1b04414 $")
 #endif
 
 struct radclient_list {
+	char const	*name;	/* name of this list */
+	char const	*server; /* virtual server associated with this client list */
+
 	/*
 	 *	FIXME: One set of trees for IPv4, and another for IPv6?
 	 */
 	rbtree_t	*trees[129]; /* for 0..128, inclusive. */
 	uint32_t       	min_prefix;
+
+	bool		parsed;
 };
 
 
@@ -145,6 +150,15 @@ RADCLIENT_LIST *client_list_init(CONF_SECTION *cs)
 
 	clients->min_prefix = 128;
 
+	/*
+	 *	Associate the "clients" list with the virtual server.
+	 */
+	if (cs && (cf_data_add(cs, "clients", clients, NULL) < 0)) {
+		ERROR("Failed to associate client list with section %s\n", cf_section_name1(cs));
+		client_list_free(clients);
+		return false;
+	}
+
 	return clients;
 }
 
@@ -161,6 +175,17 @@ bool client_add(RADCLIENT_LIST *clients, RADCLIENT *client)
 
 	if (!client) return false;
 
+	/*
+	 *	Initialize the global list, if not done already.
+	 */
+	if (!root_clients) {
+		root_clients = client_list_init(NULL);
+		if (!root_clients) {
+			ERROR("Cannot add client - failed creating client list");
+			return false;
+		}
+	}
+
 	/*
 	 *	Hack to fixup wildcard clients
 	 *
@@ -188,66 +213,85 @@ bool client_add(RADCLIENT_LIST *clients, RADCLIENT *client)
 	/*
 	 *	If the client also defines a server, do that now.
 	 */
-	if (client->defines_coa_server) if (!realm_home_server_add(client->coa_server)) return false;
+	if (client->defines_coa_server) if (!realm_home_server_add(client->coa_home_server)) return false;
 
 	/*
-	 *	If "clients" is NULL, it means add to the global list,
-	 *	unless we're trying to add it to a virtual server...
+	 *	If there's no client list, BUT there's a virtual
+	 *	server, try to add the client to the appropriate
+	 *	"clients" section for that virtual server.
 	 */
-	if (!clients) {
-		if (client->server != NULL) {
-			CONF_SECTION *cs;
-			CONF_SECTION *subcs;
+	if (!clients && client->server) {
+		CONF_SECTION *cs;
+		CONF_SECTION *subcs;
+		CONF_PAIR *cp;
+		char const *section_name;
 
-			cs = cf_section_sub_find_name2(main_config.config, "server", client->server);
-			if (!cs) {
-				ERROR("Failed to find virtual server %s", client->server);
-				return false;
-			}
+		cs = cf_section_sub_find_name2(main_config.config, "server", client->server);
+		if (!cs) {
+			ERROR("Cannot add client - virtual server %s does not exist", client->server);
+			return false;
+		}
 
-			/*
-			 *	If this server has no "listen" section, add the clients
-			 *	to the global client list.
-			 */
-			subcs = cf_section_sub_find(cs, "listen");
-			if (!subcs) {
-				DEBUG("No 'listen' section in virtual server %s.  Adding client to global client list",
-				      client->server);
-				goto global_clients;
-			}
+		/*
+		 *	If this server has no "listen" section, add the clients
+		 *	to the global client list.
+		 */
+		subcs = cf_section_sub_find(cs, "listen");
+		if (!subcs) {
+			DEBUG("No 'listen' section in virtual server %s.  Adding client to global client list",
+			      client->server);
+			goto check_list;
+		}
 
-			/*
-			 *	If the client list already exists, use that.
-			 *	Otherwise, create a new client list.
-			 */
-			clients = cf_data_find(cs, "clients");
-			if (!clients) {
-				clients = client_list_init(cs);
-				if (!clients) {
-					ERROR("Out of memory");
-					return false;
-				}
+		cp = cf_pair_find(subcs, "clients");
+		if (!cp) {
+			DEBUG("No 'clients' configuration item in first listener of virtual server %s.  Adding client to global client list",
+			      client->server);
+			goto check_list;
+		}
 
-				if (cf_data_add(cs, "clients", clients, (void (*)(void *)) client_list_free) < 0) {
-					ERROR("Failed to associate clients with virtual server %s", client->server);
-					client_list_free(clients);
-					return false;
-				}
-			}
+		/*
+		 *	Duplicate the lookup logic in common_socket_parse()
+		 *
+		 *	Explicit list given: use it.
+		 */
+		section_name = cf_pair_value(cp);
+		if (!section_name) goto check_list;
 
-		} else {
-		global_clients:
-			/*
-			 *	Initialize the global list, if not done already.
-			 */
-			if (!root_clients) {
-				root_clients = client_list_init(NULL);
-				if (!root_clients) return false;
-			}
-			clients = root_clients;
+		subcs = cf_section_sub_find_name2(main_config.config, "clients", section_name);
+		if (!subcs) {
+			subcs = cf_section_find(section_name);
+		}
+		if (!subcs) {
+			cf_log_err_cs(cs,
+				   "Failed to find clients %s {...}",
+				   section_name);
+			return false;
+		}
+
+		DEBUG("Adding client to client list %s", section_name);
+
+		/*
+		 *	If the client list already exists, use that.
+		 *	Otherwise, create a new client list.
+		 *
+		 *	@todo - add the client to _all_ listeners?
+		 */
+		clients = cf_data_find(subcs, "clients");
+		if (clients) goto check_list;
+
+		clients = client_list_init(subcs);
+		if (!clients) {
+			ERROR("Cannot add client - failed creating client list %s for server %s", section_name,
+			      client->server);
+			return false;
 		}
 	}
 
+check_list:
+	if (!clients) clients = root_clients;
+	client->list = clients;
+
 	/*
 	 *	Create a tree for it.
 	 */
@@ -280,8 +324,8 @@ bool client_add(RADCLIENT_LIST *clients, RADCLIENT *client)
 #endif
 #ifdef WITH_COA
 		    namecmp(coa_name) &&
-		    (old->coa_server == client->coa_server) &&
-		    (old->coa_pool == client->coa_pool) &&
+		    (old->coa_home_server == client->coa_home_server) &&
+		    (old->coa_home_pool == client->coa_home_pool) &&
 #endif
 		    (old->message_authenticator == client->message_authenticator)) {
 			WARN("Ignoring duplicate client %s", client->longname);
@@ -512,14 +556,34 @@ RADCLIENT_LIST *client_list_parse_section(CONF_SECTION *section, UNUSED bool tls
 	 *	it.  Otherwise create a new one.
 	 */
 	clients = cf_data_find(section, "clients");
-	if (clients) return clients;
+	if (clients) {
+		/*
+		 *	Modules are initialized before the listeners.
+		 *	Which means that we MIGHT have read clients
+		 *	from SQL before parsing this "clients"
+		 *	section.  So there may already be a clients
+		 *	list.
+		 *
+		 *	But the list isn't _our_ list that we parsed,
+		 *	so we still need to parse the clients here.
+		 */
+		if (clients->parsed) return clients;		
+	} else {
+		clients = client_list_init(section);
+		if (!clients) return NULL;
+	}
 
-	clients = client_list_init(section);
-	if (!clients) return NULL;
+	if (cf_top_section(section) == section) {
+		global = true;
+		clients->name = "global";
+		clients->server = NULL;
+	}
 
-	if (cf_top_section(section) == section) global = true;
-
-	if (strcmp("server", cf_section_name1(section)) == 0) in_server = true;
+	if (strcmp("server", cf_section_name1(section)) == 0) {
+		clients->name = NULL;
+		clients->server = cf_section_name2(section);
+		in_server = true;
+	}
 
 	for (cs = cf_subsection_find_next(section, NULL, "client");
 	     cs;
@@ -632,15 +696,6 @@ RADCLIENT_LIST *client_list_parse_section(CONF_SECTION *section, UNUSED bool tls
 
 	}
 
-	/*
-	 *	Associate the clients structure with the section.
-	 */
-	if (cf_data_add(section, "clients", clients, NULL) < 0) {
-		cf_log_err_cs(section, "Failed to associate clients with section %s", cf_section_name1(section));
-		client_list_free(clients);
-		return NULL;
-	}
-
 	/*
 	 *	Replace the global list of clients with the new one.
 	 *	The old one is still referenced from the original
@@ -648,6 +703,7 @@ RADCLIENT_LIST *client_list_parse_section(CONF_SECTION *section, UNUSED bool tls
 	 */
 	if (global) root_clients = clients;
 
+	clients->parsed = true;
 	return clients;
 }
 
@@ -824,7 +880,7 @@ int client_map_section(CONF_SECTION *out, CONF_SECTION const *map, client_value_
  * @param ctx to allocate new clients in.
  * @param cs to process as a client.
  * @param in_server Whether the client should belong to a specific virtual server.
- * @param with_coa If true and coa_server or coa_pool aren't specified automatically,
+ * @param with_coa If true and coa_home_server or coa_home_pool aren't specified automatically,
  *	create a coa home_server section and add it to the client CONF_SECTION.
  * @return new RADCLIENT struct.
  */
@@ -1054,11 +1110,11 @@ RADCLIENT *client_afrom_cs(TALLOC_CTX *ctx, CONF_SECTION *cs, bool in_server, bo
 		cp = cf_pair_find(cs, "coa_server");
 		if (cp) {
 			c->coa_name = cf_pair_value(cp);
-			c->coa_pool = home_pool_byname(c->coa_name, HOME_TYPE_COA);
-			if (!c->coa_pool) {
-				c->coa_server = home_server_byname(c->coa_name, HOME_TYPE_COA);
+			c->coa_home_pool = home_pool_byname(c->coa_name, HOME_TYPE_COA);
+			if (!c->coa_home_pool) {
+				c->coa_home_server = home_server_byname(c->coa_name, HOME_TYPE_COA);
 			}
-			if (!c->coa_pool && !c->coa_server) {
+			if (!c->coa_home_pool && !c->coa_home_server) {
 				cf_log_err_cs(cs, "No such home_server or home_server_pool \"%s\"", c->coa_name);
 				goto error;
 			}
@@ -1096,7 +1152,7 @@ RADCLIENT *client_afrom_cs(TALLOC_CTX *ctx, CONF_SECTION *cs, bool in_server, bo
 
 			rad_assert(home->type == HOME_TYPE_COA);
 
-			c->coa_server = home;
+			c->coa_home_server = home;
 			c->defines_coa_server = true;
 		}
 	}
@@ -1175,7 +1231,6 @@ RADCLIENT *client_afrom_request(RADCLIENT_LIST *clients, REQUEST *request)
 	int		i, *pi;
 	char		**p;
 	RADCLIENT	*c;
-	CONF_PAIR	*cp = NULL;
 	char		buffer[128];
 
 	vp_cursor_t	cursor;
@@ -1206,6 +1261,7 @@ RADCLIENT *client_afrom_request(RADCLIENT_LIST *clients, REQUEST *request)
 	for (i = 0; dynamic_config[i].name != NULL; i++) {
 		DICT_ATTR const *da;
 		char *strvalue = NULL;
+		CONF_PAIR *cp = NULL;
 
 		da = dict_attrbyname(dynamic_config[i].name);
 		if (!da) {
@@ -1377,6 +1433,8 @@ RADCLIENT *client_afrom_request(RADCLIENT_LIST *clients, REQUEST *request)
 	fr_cursor_first(&cursor);
 	vp = fr_cursor_remove(&cursor);
 	if (vp) {
+		CONF_PAIR *cp;
+
 		do {
 			char *value;
 
