@@ -1,7 +1,7 @@
 /*
  * parser.c	Parse various things
  *
- * Version:	$Id: d83516b035eccd7c27e871039225ae2fd8a1f8e9 $
+ * Version:	$Id: 7bafa8c70b2ab89fae50f4943b8984e5b4e48e36 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
  * Copyright 2013  Alan DeKok <aland@freeradius.org>
  */
 
-RCSID("$Id: d83516b035eccd7c27e871039225ae2fd8a1f8e9 $")
+RCSID("$Id: 7bafa8c70b2ab89fae50f4943b8984e5b4e48e36 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/parser.h>
@@ -57,11 +57,21 @@ size_t fr_cond_sprint(char *buffer, size_t bufsize, fr_cond_t const *in)
 	rad_assert(bufsize > 0);
 
 next:
+	rad_assert(p < end);
+
 	if (!c) {
 		p[0] = '\0';
 		return 0;
 	}
 
+	/*
+	 *	Don't overflow the output buffer.
+	 */
+	if ((end - p) < 2) {
+		p[0] = '\0';
+		return 0;
+	}
+
 	if (c->negate) {
 		*(p++) = '!';	/* FIXME: only allow for child? */
 	}
@@ -70,9 +80,9 @@ next:
 	case COND_TYPE_EXISTS:
 		rad_assert(c->data.vpt != NULL);
 		if (c->cast) {
-			len = snprintf(p, end - p, "<%s>", fr_int2str(dict_attr_types,
-								      c->cast->type, "??"));
-			p += len;
+			snprintf(p, end - p, "<%s>", fr_int2str(dict_attr_types,
+								c->cast->type, "??"));
+			p += strlen(p);
 		}
 
 		len = tmpl_prints(p, end - p, c->data.vpt, NULL);
@@ -85,9 +95,9 @@ next:
 		*(p++) = '[';	/* for extra-clear debugging */
 #endif
 		if (c->cast) {
-			len = snprintf(p, end - p, "<%s>", fr_int2str(dict_attr_types,
-								      c->cast->type, "??"));
-			p += len;
+			snprintf(p, end - p, "<%s>", fr_int2str(dict_attr_types,
+								c->cast->type, "??"));
+			p += strlen(p);
 		}
 
 		len = map_prints(p, end - p, c->data.map);
@@ -560,13 +570,20 @@ static ssize_t condition_tokenize(TALLOC_CTX *ctx, CONF_ITEM *ci, char const *st
 		}
 		p += slen;
 
+#ifndef __clang_analyzer__
 		lhs_p = p;
+#endif
 		slen = condition_tokenize_word(c, p, &lhs, &lhs_type, error);
 		if (slen <= 0) {
 			return_SLEN;
 		}
 		p += slen;
 
+
+#ifdef __clang_analyzer__
+		if (!lhs) return_P("Internal error");
+#endif
+
 		/*
 		 *	If the LHS is 0xabcdef... automatically cast it to octets
 		 */
@@ -746,6 +763,10 @@ static ssize_t condition_tokenize(TALLOC_CTX *ctx, CONF_ITEM *ci, char const *st
 					return_SLEN;
 				}
 
+#ifdef __clang_analyzer__
+				if (!cast_da) return_P("Internal error");
+#endif
+
 				if (!c->cast) {
 					return_P("Unexpected cast");
 				}
@@ -1145,7 +1166,6 @@ static ssize_t condition_tokenize(TALLOC_CTX *ctx, CONF_ITEM *ci, char const *st
 							 *	by modules.c, so we can't enforce strictness here.
 							 */
 							c->pass2_fixup = PASS2_FIXUP_TYPE;
-
 						} else {
 							return_rhs("Failed to parse value for attribute");
 						}
@@ -1361,6 +1381,8 @@ done:
 	 *	run-time evaluation has fewer cases to check.
 	 */
 	if (c->type == COND_TYPE_MAP) do {
+		VERIFY_MAP(c->data.map);
+
 		/*
 		 *	!FOO !~ BAR --> FOO =~ BAR
 		 */
@@ -1542,6 +1564,8 @@ done:
 	 *	"foo" is NOT the same as 'foo' or a bare foo.
 	 */
 	if (c->type == COND_TYPE_EXISTS) {
+		VERIFY_TMPL(c->data.vpt);
+
 		switch (c->data.vpt->type) {
 		case TMPL_TYPE_XLAT:
 		case TMPL_TYPE_ATTR:
