@@ -1,7 +1,7 @@
 /*
  * radattr.c	RADIUS Attribute debugging tool.
  *
- * Version:	$Id: c402f079d16598be8005a39a05345a1c65ec6e22 $
+ * Version:	$Id: cf0809c863f603f7266da3ab7bb9be494e4546f7 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
  * Copyright 2010  Alan DeKok <aland@freeradius.org>
  */
 
-RCSID("$Id: c402f079d16598be8005a39a05345a1c65ec6e22 $")
+RCSID("$Id: cf0809c863f603f7266da3ab7bb9be494e4546f7 $")
 
 #include <freeradius-devel/libradius.h>
 
@@ -65,21 +65,29 @@ static ssize_t xlat_test(UNUSED void *instance, UNUSED REQUEST *request,
 	return 0;
 }
 
-static RADIUS_PACKET my_original = {
+static RADIUS_PACKET access_request = {
 	.sockfd = -1,
 	.id = 0,
 	.code = PW_CODE_ACCESS_REQUEST,
 	.vector = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f },
 };
 
-
-static RADIUS_PACKET my_packet = {
+static RADIUS_PACKET access_accept = {
 	.sockfd = -1,
 	.id = 0,
 	.code = PW_CODE_ACCESS_ACCEPT,
 	.vector = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f },
 };
 
+static RADIUS_PACKET coa_request = {
+	.sockfd = -1,
+	.id = 0,
+	.code = PW_CODE_COA_REQUEST,
+	.vector = { 0 },
+};
+
+static RADIUS_PACKET *my_original = &access_request;
+static RADIUS_PACKET *my_packet = &access_accept;
 
 static char const *my_secret = "testing123";
 
@@ -682,6 +690,33 @@ static void process_file(const char *root_dir, char const *filename)
 			continue;
 		}
 
+		if (strncmp(p, "packet ", 7) == 0) {
+			p += 7;
+			if (strncmp(p, "access_accept", 13) == 0) {
+				my_packet = &access_accept;
+			} else if (strncmp(p, "coa_request", 11) == 0) {
+				my_packet = &coa_request;
+			} else {
+				fprintf(stderr, "Unsupported packet type at line %d of %s: %s\n",
+					lineno, directory, p);
+				exit(1);
+			}
+			continue;
+		}
+		if (strncmp(p, "original ", 9) == 0) {
+			p += 9;
+			if (strncmp(p, "null", 4) == 0) {
+				my_original = NULL;
+			} else if (strncmp(p, "access_request", 14) == 0) {
+				my_original = &access_request;
+			} else {
+				fprintf(stderr, "Unsupported original type at line %d of %s: %s\n",
+					lineno, directory, p);
+				exit(1);
+			}
+			continue;
+		}
+
 		if (strncmp(p, "encode ", 7) == 0) {
 			if (strcmp(p + 7, "-") == 0) {
 				p = output;
@@ -702,7 +737,7 @@ static void process_file(const char *root_dir, char const *filename)
 
 				memcpy(&qvp, &pvp, sizeof(pvp));
 
-				len = rad_vp2attr(&my_packet, &my_original, my_secret, qvp,
+				len = rad_vp2attr(my_packet, my_original, my_secret, qvp,
 						  attr, data + sizeof(data) - attr);
 				if (len < 0) {
 					fprintf(stderr, "Failed encoding %s: %s\n",
@@ -737,7 +772,7 @@ static void process_file(const char *root_dir, char const *filename)
 			my_len = 0;
 			while (len > 0) {
 				vp = NULL;
-				my_len = rad_attr2vp(NULL, &my_packet, &my_original, my_secret, attr, len, &vp);
+				my_len = rad_attr2vp(NULL, my_packet, my_original, my_secret, attr, len, &vp);
 				if (my_len < 0) {
 					fr_pair_list_free(&head);
 					break;
@@ -786,6 +821,7 @@ static void process_file(const char *root_dir, char const *filename)
 			continue;
 		}
 
+#ifdef WITH_DHCP
 		/*
 		 *	And some DHCP tests
 		 */
@@ -868,6 +904,7 @@ static void process_file(const char *root_dir, char const *filename)
 			}
 			continue;
 		}
+#endif
 
 		if (strncmp(p, "attribute ", 10) == 0) {
 			p += 10;
@@ -918,6 +955,65 @@ static void process_file(const char *root_dir, char const *filename)
 	if (fp != stdin) fclose(fp);
 }
 
+/** Dump all of the dictionary entries as
+ *
+ *	ALIAS name OID
+ *
+ *  To create dictionaries which allow files to be used with v4.
+ *
+ * rm -rf alias;mkdir alias;./build/make/jlibtool --mode=execute ./build/bin/radattr  -D ./share/ -A  | sort -n -k6 -k7 -k8 -k9 -k10 -k11 | gawk '{printf "%s\t%-40s\t%s\n", $1, $2, $3 >> "alias/alias." tolower($5) }'
+ *
+ *  And then post-process each file to remove the comments.
+ *
+ *  Note that we have to use GNU Awk, as OSX awk doesn't like redirection to a file which includes a variable.
+ */
+static int dump_aliases(void *ctx, void *data)
+{
+	DICT_ATTR *da = data;
+	FILE *fp = ctx;
+	int nest, attr, dv_type;
+	DICT_VENDOR *dv;
+	char buffer[1024];
+
+	if (!da->vendor || (da->vendor > FR_MAX_VENDOR)) return 0;
+
+	dv = dict_vendorbyvalue(da->vendor);
+	dv_type = dv->type;
+
+	(void) dict_print_oid(buffer, sizeof(buffer), da);
+	fprintf(fp, "ALIAS\t%s\t%s # %s %u", da->name, buffer, dv->name, da->vendor);
+
+	attr = da->attr;
+	switch (dv_type) {
+	default:
+	case 1:
+		fprintf(fp, " %u", attr & 0xff);
+
+		/*
+		 *	Only these ones are bit-packed.
+		 */
+		for (nest = 1; nest <= fr_attr_max_tlv; nest++) {
+			if (((attr >> fr_attr_shift[nest]) & fr_attr_mask[nest]) == 0) break;
+
+			fprintf(fp, " %u",
+				(attr >> fr_attr_shift[nest]) & fr_attr_mask[nest]);
+		}
+		break;
+
+	case 2:
+		fprintf(fp, " %u", attr & 0xffff);
+		break;
+
+	case 4:
+		fprintf(fp, " %u", attr);
+		break;
+	}
+
+	printf("\n");
+
+	return 0;
+}
+
 static void NEVER_RETURNS usage(void)
 {
 	fprintf(stderr, "usage: radattr [OPTS] filename\n");
@@ -933,6 +1029,7 @@ int main(int argc, char *argv[])
 {
 	int c;
 	bool report = false;
+	bool dump_alias = false;
 	char const *radius_dir = RADDBDIR;
 	char const *dict_dir = DICTDIR;
 	int *inst = &c;
@@ -946,7 +1043,10 @@ int main(int argc, char *argv[])
 	}
 #endif
 
-	while ((c = getopt(argc, argv, "d:D:xMh")) != EOF) switch (c) {
+	while ((c = getopt(argc, argv, "Ad:D:xMh")) != EOF) switch (c) {
+		case 'A':
+			dump_alias = true;
+			break;
 		case 'd':
 			radius_dir = optarg;
 			break;
@@ -990,6 +1090,11 @@ int main(int argc, char *argv[])
 		return 1;
 	}
 
+	if (dump_alias) {
+		(void) dict_walk(dump_aliases, stdout);
+		return 0;
+	}
+
 	if (argc < 2) {
 		process_file(NULL, "-");
 
