@@ -1,5 +1,5 @@
 /**
- * $Id: 8d0c826970e344578fc40fcbbdcbce2e8a90c37a $
+ * $Id: f3e066783f56dc56e6d16c6ed7562e506599b64b $
  * @file ldap.h
  * @brief LDAP authorization and authentication module headers.
  *
@@ -20,6 +20,7 @@
  *	always need to support that.
  */
 #define LDAP_DEPRECATED 1
+USES_APPLE_DEPRECATED_API	/* Apple wants us to use OpenDirectory Framework, we don't want that */
 #include <lber.h>
 #include <ldap.h>
 #include "config.h"
@@ -162,6 +163,9 @@ typedef struct ldap_instance {
 
 	int		userobj_scope;			//!< Search scope.
 
+	char const	*user_dn;			//!< for multiple LDAP modules
+	DICT_ATTR const *user_dn_da;			//!< cached user DN
+
 	char const	*userobj_membership_attr;	//!< Attribute that describes groups the user is a member of.
 	char const	*userobj_access_attr;		//!< Attribute to check to see if the user should be locked out.
 	bool		access_positive;		//!< If true the presence of the attribute will allow access,
@@ -203,6 +207,10 @@ typedef struct ldap_instance {
 	DICT_ATTR const	*group_da;			//!< The DA associated with this specific instance of the
 							//!< rlm_ldap module.
 
+	bool		allow_dangling_group_refs;	//!< Don't error if we fail to resolve a group DN referenced
+							///< from a user object.
+
+
 	/*
 	 *	Dynamic clients
 	 */
@@ -258,6 +266,9 @@ typedef struct ldap_instance {
 
 	int		tls_require_cert;		//!< OpenLDAP constant representing the require cert string.
 
+	char const	*tls_min_version_str;		//!< Minimum TLS version
+	int		tls_min_version;
+
 	/*
 	 *	Options
 	 */
@@ -301,8 +312,6 @@ typedef struct ldap_handle {
 	LDAP		*handle;			//!< libldap handle.
 	bool		rebound;			//!< Whether the connection has been rebound to something
 							//!< other than the admin user.
-	bool		referred;			//!< Whether the connection is now established a server
-							//!< other than the configured one.
 	rlm_ldap_t	*inst;				//!< rlm_ldap configuration.
 } ldap_handle_t;
 
