@@ -15,7 +15,7 @@
  */
 
 /**
- * $Id: fdd6f01ecf008933de231c22787cc0947ad9fbc8 $
+ * $Id: 33e5a885caa8dfd4069692b4b1ae8627f4592c25 $
  * @file rlm_ldap.c
  * @brief LDAP authorization and authentication module.
  *
@@ -27,7 +27,7 @@
  * @copyright 2012 Alan DeKok <aland@freeradius.org>
  * @copyright 1999-2013 The FreeRADIUS Server Project.
  */
-RCSID("$Id: fdd6f01ecf008933de231c22787cc0947ad9fbc8 $")
+RCSID("$Id: 33e5a885caa8dfd4069692b4b1ae8627f4592c25 $")
 
 #include	<freeradius-devel/rad_assert.h>
 
@@ -106,6 +106,10 @@ static CONF_PARSER tls_config[] = {
 	{ "randfile", FR_CONF_OFFSET(PW_TYPE_FILE_EXISTS | PW_TYPE_DEPRECATED, rlm_ldap_t, tls_random_file), NULL },
 	{ "random_file", FR_CONF_OFFSET(PW_TYPE_FILE_EXISTS, rlm_ldap_t, tls_random_file), NULL },
 
+#ifdef LDAP_OPT_X_TLS_PROTOCOL_MIN
+	{ "tls_min_version", FR_CONF_OFFSET(PW_TYPE_STRING, rlm_ldap_t, tls_min_version_str), NULL },
+#endif
+
 	/*
 	 *	LDAP Specific TLS attributes
 	 */
@@ -922,7 +926,7 @@ static int mod_instantiate(CONF_SECTION *conf, void *instance)
 			 */
 			if ((p = strchr(value, ']')) && (p[1] == ':')) {			/* IPv6 */
 				set_port_maybe = false;
-			} else if ((p = strchr(value, ':')) && (p = strchr(p + 1, ':'))) {	/* IPv4 */
+			} else if ((p = strchr(value, ':')) && (strchr(p + 1, ':') != NULL)) {  /* IPv4 */
 				set_port_maybe = false;
 			}
 
@@ -1188,6 +1192,30 @@ static int mod_instantiate(CONF_SECTION *conf, void *instance)
 #endif
 	}
 
+	if (inst->tls_min_version_str) {
+#ifdef LDAP_OPT_X_TLS_PROTOCOL_MIN
+		if (strcmp(inst->tls_min_version_str, "1.2") == 0) {
+			inst->tls_min_version = LDAP_OPT_X_TLS_PROTOCOL_TLS1_2;
+
+		} else if (strcmp(inst->tls_min_version_str, "1.1") == 0) {
+			inst->tls_min_version = LDAP_OPT_X_TLS_PROTOCOL_TLS1_1;
+
+		} else if (strcmp(inst->tls_min_version_str, "1.0") == 0) {
+			inst->tls_min_version = LDAP_OPT_X_TLS_PROTOCOL_TLS1_0;
+
+		} else {
+			cf_log_err_cs(conf, "Invalid 'tls.tls_min_version' value \"%s\"", inst->tls_min_version_str);
+			goto error;
+		}
+#else
+		cf_log_err_cs(conf, "This version of libldap does not support tls.tls_min_version."
+			      " Please upgrade or substitute current libldap and "
+			      "rebuild this module");
+		goto error;
+
+#endif
+	}
+
 	/*
 	 *	Build the attribute map
 	 */
