@@ -15,7 +15,7 @@
  */
 
 /**
- * $Id: 9e62e6230c722e0d690241ded5b897192597c787 $
+ * $Id: a5c3b937e8da84a872a712cf355deddc18a4dbf5 $
  * @file rlm_sql.c
  * @brief Implements FreeTDS rlm_sql driver.
  *
@@ -24,7 +24,7 @@
  * @copyright 2000  Mattias Sjostrom <mattias@nogui.se>
  */
 
-RCSID("$Id: 9e62e6230c722e0d690241ded5b897192597c787 $")
+RCSID("$Id: a5c3b937e8da84a872a712cf355deddc18a4dbf5 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/rad_assert.h>
@@ -42,6 +42,7 @@ typedef struct rlm_sql_freetds_conn {
 	char		**results;	//!< Result strings from statement execution.
 	char		*error;		//!< The last error string created by one of the call backs.
 	bool		established;	//!< Set to false once the connection has been properly established.
+	CS_INT		rows_affected;	//!< Rows affected by last INSERT / UPDATE / DELETE.
 } rlm_sql_freetds_conn_t;
 
 #define	MAX_DATASTR_LEN	256
@@ -196,6 +197,12 @@ static sql_rcode_t sql_query(rlm_sql_handle_t *handle, UNUSED rlm_sql_config_t *
 	CS_RETCODE	results_ret;
 	CS_INT		result_type;
 
+	/*
+	 *	Reset rows_affected in case the query fails.
+	 *	Prevents accidentally returning the rows_affected from a previous query.
+	 */
+	conn->rows_affected = -1;
+
 	if (ct_cmd_alloc(conn->db, &conn->command) != CS_SUCCEED) {
 		ERROR("rlm_sql_freetds: Unable to allocate command structure (ct_cmd_alloc())");
 
@@ -252,6 +259,17 @@ static sql_rcode_t sql_query(rlm_sql_handle_t *handle, UNUSED rlm_sql_config_t *
 		}
 	}
 
+	/*
+	 *	Retrieve the number of rows affected - the later calls
+	 *	to ct_results end up resetting the underlying counter so we
+	 *	no longer have access to this.
+	 */
+	if (ct_res_info(conn->command, CS_ROW_COUNT, &conn->rows_affected, CS_UNUSED, NULL) != CS_SUCCEED) {
+		ERROR("rlm_sql_freetds: error retrieving row count");
+
+		return RLM_SQL_ERROR;
+	}
+
 	/*
 	 *	Second call to ct_results, we need returncode CS_SUCCEED
 	 *	and result_type CS_CMD_DONE.
@@ -500,15 +518,8 @@ static sql_rcode_t sql_select_query(rlm_sql_handle_t *handle, rlm_sql_config_t *
 static int sql_num_rows(rlm_sql_handle_t *handle, UNUSED rlm_sql_config_t *config)
 {
 	rlm_sql_freetds_conn_t *conn = handle->conn;
-	CS_INT	num;
 
-	if (ct_res_info(conn->command, CS_ROW_COUNT, &num, CS_UNUSED, NULL) != CS_SUCCEED) {
-		ERROR("rlm_sql_freetds: error retrieving row count");
-
-		return RLM_SQL_ERROR;
-	}
-
-	return num;
+	return (conn->rows_affected);
 }
 
 static sql_rcode_t sql_fetch_row(rlm_sql_handle_t *handle, UNUSED rlm_sql_config_t *config)
@@ -575,6 +586,7 @@ static sql_rcode_t sql_finish_query(rlm_sql_handle_t *handle, UNUSED rlm_sql_con
 		return RLM_SQL_ERROR;
 	}
 	conn->command = NULL;
+	conn->rows_affected = -1;
 
 	return RLM_SQL_OK;
 }
