@@ -15,7 +15,7 @@
  */
 
 /**
- * $Id: 0c872ea24653a1b68a71a56072b4324fe06e321a $
+ * $Id: 24638f09cd9f00b3a97d832854f5babc181d12c6 $
  * @file rlm_sql.c
  * @brief Implements SQL 'users' file, and SQL accounting.
  *
@@ -24,7 +24,7 @@
  * @copyright 2000  Mike Machado <mike@innercite.com>
  * @copyright 2000  Alan DeKok <aland@ox.org>
  */
-RCSID("$Id: 0c872ea24653a1b68a71a56072b4324fe06e321a $")
+RCSID("$Id: 24638f09cd9f00b3a97d832854f5babc181d12c6 $")
 
 #include <ctype.h>
 
@@ -303,7 +303,7 @@ static int generate_sql_clients(rlm_sql_t *inst)
 
 		num_rows = (inst->module->sql_num_fields)(handle, inst->config);
 		if (num_rows < 5) {
-			WARN("SELECT returned too few rows.  Please do not edit 'client_query'");
+			WARN("SELECT returned too few fields.  Please do not edit 'client_query'");
 			continue;
 		}
 
@@ -573,7 +573,7 @@ static int sql_get_grouplist(rlm_sql_t *inst, rlm_sql_handle_t **handle, REQUEST
 
 	if (!inst->config->groupmemb_query) return 0;
 
-	if (radius_axlat(&expanded, request, inst->config->groupmemb_query, sql_escape_for_xlat_func, inst) < 0) return -1;
+	if (radius_axlat(&expanded, request, inst->config->groupmemb_query, sql_escape_func, *handle) < 0) return -1;
 
 	ret = rlm_sql_select_query(inst, request, handle, expanded);
 	talloc_free(expanded);
@@ -1532,6 +1532,20 @@ static int acct_redundant(rlm_sql_t *inst, REQUEST *request, sql_acct_section_t
 		pair = cf_pair_find_next(section->cs, pair, attr);
 
 		if (!pair) {
+			char const *name;
+
+			/*
+			 *	Hack for RADIUS!
+			 */
+			name = cf_section_name1(section->cs);
+			if ((strcmp(name, "accounting-on") == 0) ||
+			    (strcmp(name, "accounting-off") == 0)) {
+				RDEBUG("Accounting on/off had no updates.  Returning 'ok'");
+				rcode = RLM_MODULE_OK;
+				goto finish;
+			}
+
+
 			RDEBUG("No additional queries configured");
 			rcode = RLM_MODULE_NOOP;
 
@@ -1590,7 +1604,7 @@ static rlm_rcode_t mod_checksimul(void *instance, REQUEST * request)
 	char const     		*call_num = NULL;
 	VALUE_PAIR		*vp;
 	int			ret;
-	uint32_t		nas_addr = 0;
+	fr_ipaddr_t		nas_addr;
 	uint32_t		nas_port = 0;
 
 	char 			*expanded = NULL;
@@ -1713,14 +1727,18 @@ static rlm_rcode_t mod_checksimul(void *instance, REQUEST * request)
 		}
 
 		if (row[3]) {
-			nas_addr = inet_addr(row[3]);
+			if (fr_pton(&nas_addr, row[3], -1, AF_UNSPEC, false) < 0) {
+				RDEBUG("Cannot parse '%s' as an IPv4 or an IPv6 address - %s", row[3], fr_strerror());
+				rcode = RLM_MODULE_FAIL;
+				goto finish;
+			}
 		}
 
 		if (row[4]) {
 			nas_port = atoi(row[4]);
 		}
 
-		check = rad_check_ts(nas_addr, nas_port, row[2], row[1]);
+		check = rad_check_ts(&nas_addr, nas_port, row[2], row[1]);
 		if (check == 0) {
 			/*
 			 *	Stale record - zap it.
@@ -1740,7 +1758,7 @@ static rlm_rcode_t mod_checksimul(void *instance, REQUEST * request)
 				}
 				if ((num_rows > 8) && row[8])
 					sess_time = atoi(row[8]);
-				session_zap(request, nas_addr, nas_port,
+				session_zap(request, &nas_addr, nas_port,
 					    row[2], row[1], framed_addr,
 					    proto, sess_time);
 			}
