@@ -15,14 +15,14 @@
  */
 
 /**
- * $Id: 47ebfe0b3a342df5ee8fb4f22f90f20984e1e99f $
+ * $Id: 26aded7a7c58e6790271bb30e867801a81f4ab79 $
  * @file rlm_sqlcounter.c
  * @brief Tracks data usage and other counters using SQL.
  *
  * @copyright 2001,2006  The FreeRADIUS server project
  * @copyright 2001  Alan DeKok <aland@ox.org>
  */
-RCSID("$Id: 47ebfe0b3a342df5ee8fb4f22f90f20984e1e99f $")
+RCSID("$Id: 26aded7a7c58e6790271bb30e867801a81f4ab79 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -68,6 +68,7 @@ typedef struct rlm_sqlcounter_t {
 					//!< session time.
 	char const	*reset;  	//!< Daily, weekly, monthly,
 					//!< never or user defined.
+	uint32_t	reset_day;  	//!< Reset day.
 	time_t		reset_time;
 	time_t		last_reset;
 	DICT_ATTR const	*key_attr;	//!< Attribute number for key field.
@@ -92,6 +93,8 @@ static const CONF_PARSER module_config[] = {
 	{ "query", FR_CONF_OFFSET(PW_TYPE_STRING | PW_TYPE_XLAT | PW_TYPE_REQUIRED, rlm_sqlcounter_t, query), NULL },
 	{ "reset", FR_CONF_OFFSET(PW_TYPE_STRING | PW_TYPE_REQUIRED, rlm_sqlcounter_t, reset), NULL },
 
+	{ "reset_day", FR_CONF_OFFSET(PW_TYPE_INTEGER, rlm_sqlcounter_t, reset_day), "1" },
+
 	{ "counter-name", FR_CONF_OFFSET(PW_TYPE_STRING | PW_TYPE_DEPRECATED, rlm_sqlcounter_t, counter_name), NULL },
 	{ "counter_name", FR_CONF_OFFSET(PW_TYPE_STRING | PW_TYPE_REQUIRED, rlm_sqlcounter_t, counter_name), NULL },
 
@@ -111,6 +114,7 @@ static int find_next_reset(rlm_sqlcounter_t *inst, REQUEST *request, time_t time
 	char last = '\0';
 	struct tm *tm, s_tm;
 	char sCurrentTime[40], sNextTime[40];
+	bool is_monthly = false;
 
 	tm = localtime_r(&timeval, &s_tm);
 	tm->tm_sec = tm->tm_min = 0;
@@ -147,7 +151,6 @@ static int find_next_reset(rlm_sqlcounter_t *inst, REQUEST *request, time_t time
 		tm->tm_hour = 0;
 		tm->tm_mday += num;
 		inst->reset_time = mktime(tm);
-
 	} else if (strcmp(inst->reset, "weekly") == 0 || last == 'w') {
 		/*
 		 *  Round up to the next nearest week.
@@ -158,10 +161,10 @@ static int find_next_reset(rlm_sqlcounter_t *inst, REQUEST *request, time_t time
 
 	} else if (strcmp(inst->reset, "monthly") == 0 || last == 'm') {
 		tm->tm_hour = 0;
-		tm->tm_mday = 1;
+		tm->tm_mday = inst->reset_day;
 		tm->tm_mon += num;
 		inst->reset_time = mktime(tm);
-
+		is_monthly = true;
 	} else if (strcmp(inst->reset, "never") == 0) {
 		inst->reset_time = 0;
 
@@ -176,9 +179,14 @@ static int find_next_reset(rlm_sqlcounter_t *inst, REQUEST *request, time_t time
 
 	len = strftime(sNextTime, sizeof(sNextTime),"%Y-%m-%d %H:%M:%S",tm);
 	if (len == 0) *sNextTime = '\0';
-	RDEBUG2("rlm_sqlcounter: Current Time: %" PRId64 " [%s], Next reset %" PRId64 " [%s]",
-		(int64_t) timeval, sCurrentTime, (int64_t) inst->reset_time, sNextTime);
 
+	if (is_monthly) {
+		DEBUG("rlm_sqlcounter: Current Time: %" PRId64 " [%s], Next reset %" PRId64 " [%s], Reset day [%d]",
+			(int64_t) timeval, sCurrentTime, (int64_t) inst->reset_time, sNextTime, inst->reset_day);
+	} else {
+		DEBUG("rlm_sqlcounter: Current Time: %" PRId64 " [%s], Next reset %" PRId64 " [%s]",
+			(int64_t) timeval, sCurrentTime, (int64_t) inst->reset_time, sNextTime);
+	}
 	return ret;
 }
 
@@ -195,6 +203,7 @@ static int find_prev_reset(rlm_sqlcounter_t *inst, time_t timeval)
 	char last = '\0';
 	struct tm *tm, s_tm;
 	char sCurrentTime[40], sPrevTime[40];
+	bool is_monthly = false;
 
 	tm = localtime_r(&timeval, &s_tm);
 	len = strftime(sCurrentTime, sizeof(sCurrentTime), "%Y-%m-%d %H:%M:%S", tm);
@@ -213,6 +222,7 @@ static int find_prev_reset(rlm_sqlcounter_t *inst, time_t timeval)
 		num = atoi(inst->reset);
 		DEBUG("rlm_sqlcounter: num=%d, last=%c",num,last);
 	}
+
 	if (strcmp(inst->reset, "hourly") == 0 || last == 'h') {
 		/*
 		 *  Round down to the prev nearest hour.
@@ -238,9 +248,10 @@ static int find_prev_reset(rlm_sqlcounter_t *inst, time_t timeval)
 
 	} else if (strcmp(inst->reset, "monthly") == 0 || last == 'm') {
 		tm->tm_hour = 0;
-		tm->tm_mday = 1;
+		tm->tm_mday = inst->reset_day;
 		tm->tm_mon -= num - 1;
 		inst->last_reset = mktime(tm);
+		is_monthly = true;
 
 	} else if (strcmp(inst->reset, "never") == 0) {
 		inst->reset_time = 0;
@@ -248,22 +259,29 @@ static int find_prev_reset(rlm_sqlcounter_t *inst, time_t timeval)
 	} else {
 		return -1;
 	}
+
 	len = strftime(sPrevTime, sizeof(sPrevTime), "%Y-%m-%d %H:%M:%S", tm);
 	if (len == 0) *sPrevTime = '\0';
-	DEBUG2("rlm_sqlcounter: Current Time: %" PRId64 " [%s], Prev reset %" PRId64 " [%s]",
-	       (int64_t) timeval, sCurrentTime, (int64_t) inst->last_reset, sPrevTime);
+
+	if (is_monthly) {
+		DEBUG2("rlm_sqlcounter: Current Time: %" PRId64 " [%s], Prev reset %" PRId64 " [%s], Reset day [%d]",
+		       (int64_t) timeval, sCurrentTime, (int64_t) inst->last_reset, sPrevTime, inst->reset_day);
+	} else {
+		DEBUG2("rlm_sqlcounter: Current Time: %" PRId64 " [%s], Prev reset %" PRId64 " [%s]",
+		       (int64_t) timeval, sCurrentTime, (int64_t) inst->last_reset, sPrevTime);
+	}
 
 	return ret;
 }
 
 
 /*
- *	Replace %<whatever> in a string.
+ *	Replace %%<whatever> in a string.
  *
- *	%b	last_reset
- *	%e	reset_time
- *	%k	key_name
- *	%S	sqlmod_inst
+ *	%%b	last_reset
+ *	%%e	reset_time
+ *	%%r	reset_day
+ *	%%k	key_name
  *
  */
 
@@ -329,6 +347,12 @@ static size_t sqlcounter_expand(char *out, int outlen, char const *fmt, rlm_sqlc
 				q += strlen(q);
 				p++;
 				break;
+			case 'r': /* reset_day */
+				snprintf(tmpdt, sizeof(tmpdt), "%" PRId64, (int64_t) inst->reset_day);
+				strlcpy(q, tmpdt, freespace);
+				q += strlen(q);
+				p++;
+				break;
 			case 'k': /* Key Name */
 				WARN("Please replace '%%k' with '${key}'");
 				strlcpy(q, inst->key_name, freespace);
@@ -482,6 +506,11 @@ static int mod_instantiate(CONF_SECTION *conf, void *instance)
 	now = time(NULL);
 	inst->reset_time = 0;
 
+	if (!(inst->reset_day >= 1 && inst->reset_day <= 28)) {
+		cf_log_err_cs(conf, "Invalid reset_day '%d', valid range is from 1 to 28", inst->reset_day);
+		return -1;
+	}
+
 	if (find_next_reset(inst, NULL, now) < 0) {
 		cf_log_err_cs(conf, "Invalid reset '%s'", inst->reset);
 		return -1;
@@ -510,7 +539,7 @@ static rlm_rcode_t CC_HINT(nonnull) mod_authorize(void *instance, REQUEST *reque
 {
 	rlm_sqlcounter_t *inst = instance;
 	int rcode = RLM_MODULE_NOOP;
-	uint64_t counter, res;
+	uint64_t counter, unused;
 	DICT_ATTR const *da;
 	VALUE_PAIR *key_vp, *limit;
 	VALUE_PAIR *reply_item;
@@ -534,8 +563,8 @@ static rlm_rcode_t CC_HINT(nonnull) mod_authorize(void *instance, REQUEST *reque
 	}
 
 	/*
-	 *      Look for the key.  User-Name is special.  It means
-	 *      The REAL username, after stripping.
+	 *	Look for the key.  User-Name is special.  It means
+	 *	The REAL username, after stripping.
 	 */
 	if ((inst->key_attr->vendor == 0) && (inst->key_attr->attr == PW_USER_NAME)) {
 		key_vp = request->username;
@@ -548,7 +577,7 @@ static rlm_rcode_t CC_HINT(nonnull) mod_authorize(void *instance, REQUEST *reque
 	}
 
 	/*
-	 *      Look for the check item
+	 *	Look for the check item
 	 */
 	if ((da = dict_attrbyname(inst->limit_name)) == NULL) {
 		return rcode;
@@ -580,7 +609,6 @@ static rlm_rcode_t CC_HINT(nonnull) mod_authorize(void *instance, REQUEST *reque
 	if (radius_axlat(&expanded, request, query, NULL, NULL) < 0) {
 		return RLM_MODULE_FAIL;
 	}
-	talloc_free(expanded);
 
 	if (sscanf(expanded, "%" PRIu64, &counter) != 1) {
 		RDEBUG2("No integer found in result string \"%s\".  May be first session, setting counter to 0",
@@ -588,6 +616,8 @@ static rlm_rcode_t CC_HINT(nonnull) mod_authorize(void *instance, REQUEST *reque
 		counter = 0;
 	}
 
+	talloc_free(expanded);
+
 	/*
 	 *	Check if check item > counter
 	 */
@@ -603,7 +633,7 @@ static rlm_rcode_t CC_HINT(nonnull) mod_authorize(void *instance, REQUEST *reque
 		return RLM_MODULE_REJECT;
 	}
 
-	res = limit->vp_integer64 - counter;
+	unused = limit->vp_integer64 - counter;
 	RDEBUG2("Allowing user, &control:%s value (%" PRIu64 ") is greater than counter value (%" PRIu64 ")",
 		inst->limit_name, limit->vp_integer64, counter);
 	/*
@@ -619,12 +649,14 @@ static rlm_rcode_t CC_HINT(nonnull) mod_authorize(void *instance, REQUEST *reque
 	 *	again.  Do this only for Session-Timeout.
 	 */
 	if (((inst->reply_attr->vendor == 0) && (inst->reply_attr->attr == PW_SESSION_TIMEOUT)) &&
-	    inst->reset_time && (res >= (uint64_t)(inst->reset_time - request->timestamp))) {
-	    	uint64_t to_reset = inst->reset_time - request->timestamp;
+	    inst->reset_time && (unused >= (uint64_t)(inst->reset_time - request->timestamp))) {
+		uint64_t next_reset = inst->reset_time - request->timestamp;
 
-		RDEBUG2("Time remaining (%" PRIu64 "s) is greater than time to reset (%" PRIu64 "s).  "
-			"Adding %" PRIu64 "s to reply value", to_reset, res, to_reset);
-		res = to_reset + limit->vp_integer;
+		RDEBUG2("Time (%" PRIu64 "s) to next reset is smaller than time remaining this reset (%" PRIu64 "s).  "
+			"Extending limit to end of next reset time (%" PRIu64 "s).",
+			next_reset, unused, next_reset + limit->vp_integer);
+
+		unused = next_reset + limit->vp_integer;
 	}
 
 	/*
@@ -632,7 +664,7 @@ static rlm_rcode_t CC_HINT(nonnull) mod_authorize(void *instance, REQUEST *reque
 	 */
 	reply_item = fr_pair_find_by_da(request->reply->vps, inst->reply_attr, TAG_ANY);
 	if (reply_item) {
-		if (reply_item->vp_integer64 <= res) {
+		if (reply_item->vp_integer64 <= unused) {
 			RDEBUG2("Leaving existing &reply:%s value of %" PRIu64, inst->reply_attr->name,
 				reply_item->vp_integer64);
 
@@ -642,7 +674,7 @@ static rlm_rcode_t CC_HINT(nonnull) mod_authorize(void *instance, REQUEST *reque
 		reply_item = radius_pair_create(request->reply, &request->reply->vps, inst->reply_attr->attr,
 					       inst->reply_attr->vendor);
 	}
-	reply_item->vp_integer64 = res;
+	reply_item->vp_integer64 = unused;
 
 	RDEBUG2("Setting &reply:%s value to %" PRIu64, inst->reply_name, reply_item->vp_integer64);
 
